/* d88image.c version 1.2.5 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d88image.h"



static struct d88_format_param_t {
	unsigned char disk_type;
	int num_tracks;
	int num_sides;
	int num_sectors;
	unsigned char H_base;
	unsigned char R_base;
	unsigned char N;
	unsigned char density;
} d88_format_params[] = {
	{ D88_DISK_TYPE_2D , 40, 2, 16, 0x00, 0x01, 0x01, 0x00 },		/* 2D			*/
	{ D88_DISK_TYPE_2DD, 80, 2,  8, 0x00, 0x01, 0x02, 0x00 },		/* 2DD/8		*/
	{ D88_DISK_TYPE_2DD, 80, 2,  9, 0x00, 0x01, 0x02, 0x00 },		/* 2DD/9		*/
	{ D88_DISK_TYPE_2HD, 77, 2,  8, 0x00, 0x01, 0x03, 0x00 },		/* 2HD			*/
	{ D88_DISK_TYPE_2HD, 80, 2,  8, 0x00, 0x01, 0x03, 0x00 },		/* 2HDA			*/
	{ D88_DISK_TYPE_2HD, 80, 2,  9, 0x00, 0x0A, 0x03, 0x00 },		/* 2HS			*/
	{ D88_DISK_TYPE_2HD, 80, 2,  9, 0x80, 0x01, 0x03, 0x00 },		/* 2HDE			*/
	{ D88_DISK_TYPE_2HD, 80, 2, 15, 0x00, 0x01, 0x02, 0x00 },		/* 2HC			*/
	{ D88_DISK_TYPE_2HD, 80, 2, 18, 0x00, 0x01, 0x02, 0x00 },		/* 2HQ			*/
	{ D88_DISK_TYPE_2HD, 77, 2, 26, 0x00, 0x01, 0x01, 0x00 }		/* OS9/68000	*/
};



static int get_int2(const unsigned char *data);
static int get_int4(const unsigned char *data);
static unsigned char *put_int2(unsigned char *data, int val);
static unsigned char *put_int4(unsigned char *data, int val);
static int read_d88_track_from_buffer(struct d88_track_t *track, const unsigned char *data, int offset, int data_size);
static int read_d88_sector_from_buffer(struct d88_sector_t *sector, const unsigned char *data, int offset, int data_size);



static int get_int2(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;

	return val;
}

static int get_int4(const unsigned char *data)
{
	int val;

	val = (int)*data++;
	val |= (int)(*data++) << 8;
	val |= (int)(*data++) << 16;
	val |= (int)(*data++) << 24;

	return val;
}

static unsigned char *put_int2(unsigned char *data, int val)
{
	*data++ = val & 0xFF;
	val >>= 8;
	*data++ = val & 0xFF;

	return data;
}

static unsigned char *put_int4(unsigned char *data, int val)
{
	*data++ = val & 0xFF;
	val >>= 8;
	*data++ = val & 0xFF;
	val >>= 8;
	*data++ = val & 0xFF;
	val >>= 8;
	*data++ = val & 0xFF;

	return data;
}

void free_d88_disk(struct d88_disk_t *disk, int free_all)
{
	int t;
	int s;

	if (disk == NULL) {
		return;
	}

	for (t = 0; t < D88_TRACK_COUNT; t++) {
		if (disk->tracks[t].sectors != NULL) {
			for (s = 0; s < disk->tracks[t].num_sectors; s++) {
				if (disk->tracks[t].sectors[s].data != NULL) {
					free(disk->tracks[t].sectors[s].data);
				}
			}
			free(disk->tracks[t].sectors);
		}
	}

	if (free_all) {
		free(disk);
	} else {
		memset(disk, 0, sizeof(struct d88_disk_t));
	}
}

void free_d88_image(struct d88_image_t *image, int free_all)
{
	if (image == NULL) {
		return;
	}

	while (image->disks != NULL) {
		struct d88_disk_t *disk = image->disks;
		disk->next = NULL;
		image->disks = image->disks->next;
		free_d88_disk(disk, 1);
	}

	if (free_all) {
		free(image);
	} else {
		memset(image, 0, sizeof(struct d88_image_t));
	}
}

struct d88_disk_t *get_d88_disk(struct d88_image_t *image, int index)
{
	struct d88_disk_t *disk;

	if (index < 0) {
		return NULL;
	}

	for (disk = image->disks; disk != NULL; disk = disk->next) {
		if (index-- == 0) {
			return disk;
		}
	}

	return NULL;
}

int read_d88_image(struct d88_image_t *image, const char *name)
{
	int result = -1;

	FILE *fp = NULL;
	unsigned char *data = NULL;
	int data_size;

	if ((fp = fopen(name, "rb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("seek error(SEEK_END) '%s'\n", name);
		goto ERROR_EXIT;
	}
	data_size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("seek error(SEEK_SET) '%s'\n", name);
		goto ERROR_EXIT;
	}

	data = (unsigned char *)malloc(data_size);
	if (data == NULL) {
		printf("memory exhaust!\n");
		goto ERROR_EXIT;
	}

	if (fread(data, data_size, 1, fp) != 1) {
		printf("read error '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = read_d88_image_from_buffer(image, data, data_size);
	if (result < 0) {
		printf("read image data error '%s'\n", name);
	}

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (data != NULL) {
		free(data);
	}

	return result;
}

int read_d88_image_from_buffer(struct d88_image_t *image, const unsigned char *data, int data_size)
{
	struct d88_disk_t *last_disk = NULL;

	if (image == NULL || data == NULL || data_size <= 0) {
		printf("read_d88_image_from_buffer: illegal parameter\n");
		return -1;
	}

	memset(image, 0, sizeof(struct d88_image_t));

	while (data_size) {
		struct d88_disk_t *disk;

		if (data_size < D88_DISK_HEADER_SIZE && data_size != D88_DISK_HEADER_SIZE_OLD) {
			/* garbage ? */
			printf("data size too short ... garbage? ... ignored\n");
			break;
		}

		disk = (struct d88_disk_t *)malloc(sizeof(struct d88_disk_t));
		if (disk == NULL) {
			printf("memory exhaust!\n");
			goto ERROR_EXIT;
		}

		if (read_d88_disk_from_buffer(disk, data, data_size) < 0) {
			printf("read disk %d error\n", image->num_disks + 1);
			goto ERROR_EXIT;
		}

		data += disk->disk_size;
		data_size -= disk->disk_size;

		if (last_disk == NULL) {
			image->disks = disk;
		} else {
			last_disk->next = disk;
		}
		last_disk = disk;
		image->num_disks++;
	}

	return image->num_disks;

ERROR_EXIT:
	free_d88_image(image, 0);

	return -1;
}

int read_d88_disk_from_buffer(struct d88_disk_t *disk, const unsigned char *data, int data_size)
{
	const unsigned char *data_save = data;
	int track_count = (data_size == D88_DISK_HEADER_SIZE_OLD) ? D88_TRACK_COUNT_OLD : D88_TRACK_COUNT;
	int i;

	if (disk == NULL || data == NULL || data_size <= 0) {
		printf("read_d88_disk_from_buffer: illegal parameter\n");
		return -1;
	}

	memset(disk, 0, sizeof(struct d88_disk_t));

	memcpy(disk->name, data, D88_DISK_NAME_SIZE);
	data += D88_DISK_NAME_SIZE;
	memcpy(disk->reserved, data, D88_DISK_RESERVED_SIZE);
	data += D88_DISK_RESERVED_SIZE;
	disk->write_protect_flag = *data++;
	disk->disk_type = *data++;
	disk->disk_size = get_int4(data);
	data += 4;

	if (disk->disk_size < D88_DISK_HEADER_SIZE_OLD || data_size < disk->disk_size) {
		printf("illegal data: [%s] disk size = 0x%08X\n", disk->name, disk->disk_size);
		return -1;
	}

	if (disk->write_protect_flag != 0x00 &&
		disk->write_protect_flag != 0x10) {
		printf("illegal data: [%s] write protect flag = 0x%02X\n", disk->name, (int)disk->write_protect_flag);
		return -1;
	}

	if (disk->disk_type != 0x00 &&
		disk->disk_type != 0x10 &&
		disk->disk_type != 0x20) {
		printf("illegal data: [%s] disk type = 0x%02X\n", disk->name, (int)disk->disk_type);
		return -1;
	}

	for (i = 0; i < track_count; i++) {
		disk->track_offset[i] = get_int4(data);
		data += 4;
	}
	for (i = track_count; i < D88_TRACK_COUNT; i++) {
		disk->track_offset[i] = 0;
	}

	for (i = 0; i < D88_TRACK_COUNT_OLD; i++) {
		if (disk->track_offset[i] == 0) {
			continue;
		}
		if (disk->disk_size <= disk->track_offset[i]) {
			printf("illegal data: [%s] track %d offset = 0x%08X\n", disk->name, i, disk->track_offset[i]);
			return -1;
		} else if (disk->track_offset[i] == D88_DISK_HEADER_SIZE_OLD) {
			track_count = D88_TRACK_COUNT_OLD;
		} else if (disk->track_offset[i] < D88_DISK_HEADER_SIZE) {
			printf("illegal data: [%s] track %d offset = 0x%08X\n", disk->name, i, disk->track_offset[i]);
			return -1;
		}
	}

	if (track_count == D88_TRACK_COUNT_OLD) {
		for (i = D88_TRACK_COUNT_OLD; i < D88_TRACK_COUNT; i++) {
			disk->track_offset[i] = 0;
		}
	} else {
		for (i = D88_TRACK_COUNT_OLD; i < D88_TRACK_COUNT; i++) {
			if (disk->track_offset[i] == 0) {
				continue;
			}
			if (disk->track_offset[i] < D88_DISK_HEADER_SIZE || disk->disk_size <= disk->track_offset[i]) {
				printf("illegal data: [%s] track %d offset = 0x%08X\n", disk->name, i, disk->track_offset[i]);
				return -1;
			}
		}
	}

	for (i = 0; i < D88_TRACK_COUNT; i++) {
#if 1	/* accept unformat track */
		int track_end = disk->disk_size;
		int j;
#endif

		if (disk->track_offset[i] == 0) {
			continue;
		}

#if 1	/* accept unformat track */
		for (j = 0; j < D88_TRACK_COUNT; j++) {
			if (i == j) {
				continue;
			}
			if (disk->track_offset[i] < disk->track_offset[j] &&
				disk->track_offset[j] < track_end) {
				track_end = disk->track_offset[j];
			}
		}

		if (read_d88_track_from_buffer(disk->tracks + i,
									   data_save,
									   disk->track_offset[i],
									   track_end) < 0) {
			printf("read track %d error [%s]\n", i, disk->name);
			return -1;
		}
#else
		if (read_d88_track_from_buffer(disk->tracks + i,
									   data_save,
									   disk->track_offset[i],
									   disk->disk_size) < 0) {
			printf("read track %d error [%s]\n", i, disk->name);
			return -1;
		}
#endif
	}

	return 0;
}

static int read_d88_track_from_buffer(struct d88_track_t *track, const unsigned char *data, int offset, int data_size)
{
	int i;

	if (track == NULL || data == NULL || offset <= 0 || data_size <= 0 || data_size <= offset) {
		printf("read_d88_track_from_buffer: illegal parameter\n");
		return -1;
	}

	memset(track, 0, sizeof(struct d88_track_t));

	track->num_sectors = get_int2(data + offset + 4);
#if 1	/* accept unformat track */
	if (track->num_sectors == 0) {
		int is_unformat = 1;
		for (i = 0; i < D88_SECTOR_HEADER_SIZE; i++) {
			if (data[offset + i] != 0) {
				is_unformat = 0;
				break;
			}
		}
		if (is_unformat) {
			/* unformat track */
			track->sectors = NULL;
			return 0;
		}
	}
#endif
	if (track->num_sectors <= 0) {
		printf("illegal data: # of sectors = %d\n", track->num_sectors);
		return -1;
	}

	track->sectors = (struct d88_sector_t *)malloc(sizeof(struct d88_sector_t) * track->num_sectors);
	if (track->sectors == NULL) {
		printf("memory exhaust!\n");
		return -1;
	}

	for (i = 0; i < track->num_sectors; i++) {
		if (read_d88_sector_from_buffer(track->sectors + i, data, offset, data_size) < 0) {
			printf("read sector %d error\n", i);
			return -1;
		}

		offset += D88_SECTOR_HEADER_SIZE + track->sectors[i].sector_size;
	}

	return 0;
}

static int read_d88_sector_from_buffer(struct d88_sector_t *sector, const unsigned char *data, int offset, int data_size)
{
	if (sector == NULL || data == NULL || offset <= 0 || data_size <= 0 || data_size <= offset) {
		printf("read_d88_sector_from_buffer: illegal parameter\n");
		return -1;
	}

	memset(sector, 0, sizeof(struct d88_sector_t));

	if (data_size < offset + D88_SECTOR_HEADER_SIZE) {
		printf("data error: data size(0x%08X) < offset(0x%08X)\n", data_size, offset);
		return -1;
	}

	data += offset;

	sector->C = *data++;
	sector->H = *data++;
	sector->R = *data++;
	sector->N = *data++;
	sector->num_sectors = get_int2(data);
	data += 2;
	sector->density = *data++;
	sector->deleted_data = *data++;
	sector->status = *data++;
	memcpy(sector->reserved, data, D88_SECTOR_RESERVED_SIZE);
	data += D88_SECTOR_RESERVED_SIZE;
	sector->sector_size = get_int2(data);
	data += 2;

	if (data_size < offset + D88_SECTOR_HEADER_SIZE + sector->sector_size) {
		printf("data error: data size(0x%08X) < offset(0x%08X) + 0x10 + sector size(0x%04X)\n", data_size, offset, sector->sector_size);
		return -1;
	}

	if (sector->sector_size) {
		sector->data = (unsigned char *)malloc(sector->sector_size);
		if (sector->data == NULL) {
			printf("memory exhaust!\n");
			return -1;
		}
		memcpy(sector->data, data, sector->sector_size);
	}

	return 0;
}

int write_d88_image(const struct d88_image_t *image, const char *name)
{
	int result = -1;

	FILE *fp = NULL;
	unsigned char *data = NULL;
	int data_size;

	if (write_d88_image_to_buffer(image, &data, &data_size) < 0) {
		printf("write image data error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if ((fp = fopen(name, "wb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (fwrite(data, data_size, 1, fp) != 1) {
		printf("write error '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (data != NULL) {
		free(data);
	}

	return result;
}

int write_d88_disk(const struct d88_disk_t *disk, const char *name)
{
#if 0
	struct d88_image_t image;

	image.num_disks = 1;
	image.disks = (struct d88_disk_t *)disk;

	return write_d88_image((const struct d88_image_t *)&image, name);
#else
	const struct d88_image_t image = { 1, disk };

	return write_d88_image(&image, name);
#endif
}

int write_d88_image_to_buffer(const struct d88_image_t *image, unsigned char **data, int *data_size)
{
	int size;
	const struct d88_disk_t *disk = NULL;
	unsigned char *ptr;
	const unsigned char *data_end;
	int i;

	if (image == NULL || data == NULL) {
		printf("write_d88_image_to_buffer: illegal parameter\n");
		return -1;
	}

	*data_size = size = get_d88_image_size(image);
	*data = ptr = (unsigned char *)malloc(size);
	if (*data == NULL) {
		printf("memory exhaust!\n");
		return -1;
	}
	memset(*data, 0, size);

	data_end = ptr + size;
	for (i = 0, disk = image->disks; i < image->num_disks && disk != NULL; i++, disk = disk->next) {
		if (write_d88_disk_to_buffer(disk, ptr, &ptr, data_end - ptr) < 0) {
			printf("write d88 disk[%d] error\n", i);
			goto ERROR_EXIT;
		}
	}

	return 0;

ERROR_EXIT:
	if (*data != NULL) {
		free(*data);
	}

	return -1;
}

int write_d88_disk_to_buffer(const struct d88_disk_t *disk, unsigned char *data, unsigned char **end, int data_size)
{
	unsigned char *head = data;
	int size;
	int t;

	size = get_d88_disk_size(disk);
	if (data_size < size) {
		printf("write d88 disk: buffer size error: %d < %d\n", data_size, size);
		return -1;
	}

	memcpy(data, disk->name, D88_DISK_NAME_SIZE);
	data += D88_DISK_NAME_SIZE;
	memcpy(data, disk->reserved, D88_DISK_RESERVED_SIZE);
	data += D88_DISK_RESERVED_SIZE;
	*data++ = disk->write_protect_flag;
	*data++ = disk->disk_type;
	data = put_int4(data, size);

	data = head + D88_DISK_HEADER_SIZE;

	for (t = 0; t < D88_TRACK_COUNT; t++) {
		const struct d88_track_t *track = disk->tracks + t;
		int s;

		put_int4(head + 0x20 + t * 4, (track->sectors == NULL ? 0 : (data - head)));
		if (track->sectors == NULL) {
			continue;
		}

		for (s = 0; s < track->num_sectors; s++) {
			const struct d88_sector_t *sector = track->sectors + s;

			*data++ = sector->C;
			*data++ = sector->H;
			*data++ = sector->R;
			*data++ = sector->N;
			data = put_int2(data, track->num_sectors);
			*data++ = sector->density;
			*data++ = sector->deleted_data;
			*data++ = sector->status;
			memcpy(data, sector->reserved, D88_SECTOR_RESERVED_SIZE);
			data += D88_SECTOR_RESERVED_SIZE;
			data = put_int2(data, sector->sector_size);

			memcpy(data, sector->data, sector->sector_size);
			data += sector->sector_size;
		}
	}

	*end = data;

	return 0;
}

struct d88_image_t *create_d88_empty_image(void)
{
	struct d88_image_t *image;

	image = (struct d88_image_t *)malloc(sizeof(struct d88_image_t));
	if (image == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}

	memset(image, 0, sizeof(struct d88_image_t));
	image->num_disks = 0;
	image->disks = NULL;

	return image;
}

struct d88_disk_t *create_d88_disk(const char *name, int format)
{
	struct d88_disk_t *disk;
	int c;
	int h;
	int t;

	if (format != D88_FORMAT_2D &&
		format != D88_FORMAT_2DD_8 &&
		format != D88_FORMAT_2DD_9 &&
		format != D88_FORMAT_2HD &&
		format != D88_FORMAT_2HDA &&
		format != D88_FORMAT_2HS &&
		format != D88_FORMAT_2HDE &&
		format != D88_FORMAT_2HC &&
		format != D88_FORMAT_2HQ &&
		format != D88_FORMAT_OS9) {
		printf("create_d88_disk: illegal parameter\n");
		return NULL;
	}

	disk = (struct d88_disk_t *)malloc(sizeof(struct d88_disk_t));
	if (disk == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}
	memset(disk, 0, sizeof(struct d88_disk_t));

	if (name != NULL) {
		memcpy(disk->name, name, D88_DISK_NAME_SIZE - 1);
	}
	{
		int size = D88_DISK_NAME_SIZE - 1;
		char *des = disk->name;

		if (name != NULL) {
			const char *src = name;

			while (*src != 0 && 0 < size--) {
				*des++ = *src++;
			}
		}
		while (0 < size--) {
			*des++ = 0x20;
		}
	}

	disk->disk_type = d88_format_params[format].disk_type;

	t = 0;
	for (c = 0; c < d88_format_params[format].num_tracks; c++) {
		for (h = 0; h < d88_format_params[format].num_sides; h++) {
			disk->tracks[t].num_sectors = d88_format_params[format].num_sectors;
			disk->tracks[t].sectors = create_d88_sectors(c,
														 d88_format_params[format].H_base + h,
														 d88_format_params[format].R_base,
														 d88_format_params[format].N,
														 d88_format_params[format].num_sectors,
														 d88_format_params[format].density);
			if (disk->tracks[t].sectors == NULL) {
				printf("create sector error\n");
				free_d88_disk(disk, 1);
				return NULL;
			}
			if (t == 0) {
				if (format == D88_FORMAT_2HS) {
					disk->tracks[0].sectors[0].R = 0x01;
				} else if (format == D88_FORMAT_2HDE) {
					disk->tracks[0].sectors[0].H = 0x00;
				}
			}
			t++;
		}
	}
	disk->disk_size = get_d88_disk_size(disk);

	return disk;
}

struct d88_sector_t *create_d88_sectors(unsigned char C, unsigned char H, unsigned char R_base, unsigned char N, int num_sectors, unsigned char density)
{
	const int sector_size[6] = { 0x100, 0x200, 0x400, 0x800, 0x1000, 0x2000 };
	struct d88_sector_t *sectors;
	int size;
	int s;

	if ((D88_TRACK_COUNT / 2) <= C ||
		(H != 0x00 && H != 0x01 && H != 0x80 && H != 0x81) ||
		(N < 0x01 || 0x06 < N) ||
		num_sectors <= 0 ||
		(density != 0x00 && density != 0x40)) {
		printf("create_d88_sectors: illegal parameter\n");
		return NULL;
	}

	size = sizeof(struct d88_sector_t) * num_sectors;
	sectors = (struct d88_sector_t *)malloc(size);
	if (sectors == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}
	memset(sectors, 0, size);

	for (s = 0; s < num_sectors; s++) {
		sectors[s].C = C;
		sectors[s].H = H;
		sectors[s].R = R_base + s;
		sectors[s].N = N;
		sectors[s].num_sectors = num_sectors;
		sectors[s].density = density;
		sectors[s].deleted_data = 0x00;
		sectors[s].status = 0x00;
		sectors[s].sector_size = sector_size[N - 1];

		sectors[s].data = (unsigned char *)malloc(sectors[s].sector_size);
		if (sectors[s].data == NULL) {
			printf("memory exhaust!\n");
			return NULL;
		}
		memset(sectors[s].data, 0xE5, sectors[s].sector_size);
	}

	return sectors;
}

struct d88_sector_t *create_d88_custom_sectors(unsigned char C, unsigned char H, const unsigned char *R, const unsigned char *N, int num_sectors, unsigned char density)
{
	const int sector_size[6] = { 0x100, 0x200, 0x400, 0x800, 0x1000, 0x2000 };
	struct d88_sector_t *sectors;
	int size;
	int s;

	if ((D88_TRACK_COUNT / 2) <= C ||
		(H != 0x00 && H != 0x01 && H != 0x80 && H != 0x81) ||
		(N == NULL || num_sectors <= 0) ||
		(density != 0x00 && density != 0x40)) {
		printf("create_d88_custom_sectors: illegal parameter\n");
		return NULL;
	}

	for (s = 0; s < num_sectors; s++) {
		if (N[s] < 0x01 || 0x06 < N[s]) {
			printf("create_d88_custom_sectors: illegal parameter\n");
			return NULL;
		}
	}

	size = sizeof(struct d88_sector_t) * num_sectors;
	sectors = (struct d88_sector_t *)malloc(size);
	if (sectors == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}
	memset(sectors, 0, size);

	for (s = 0; s < num_sectors; s++) {
		sectors[s].C = C;
		sectors[s].H = H;
		sectors[s].R = (R != NULL ? R[s] : (s + 1));
		sectors[s].N = N[s];
		sectors[s].num_sectors = num_sectors;
		sectors[s].density = density;
		sectors[s].deleted_data = 0x00;
		sectors[s].status = 0x00;
		sectors[s].sector_size = sector_size[N[s] - 1];

		sectors[s].data = (unsigned char *)malloc(sectors[s].sector_size);
		if (sectors[s].data == NULL) {
			printf("memory exhaust!\n");
			return NULL;
		}
		memset(sectors[s].data, 0xE5, sectors[s].sector_size);
	}

	return sectors;
}

int get_d88_image_size(const struct d88_image_t *image)
{
	int size = 0;
	const struct d88_disk_t *disk;

	for (disk = image->disks; disk != NULL; disk = disk->next) {
		size += get_d88_disk_size(disk);
	}

	return size;
}

int get_d88_disk_size(const struct d88_disk_t *disk)
{
	int size;
	int t;

	size = D88_DISK_HEADER_SIZE;

	for (t = 0; t < D88_TRACK_COUNT; t++) {
		const struct d88_track_t *track = disk->tracks + t;
		int s;

		if (track->sectors == NULL) {
			continue;
		}

		for (s = 0; s < track->num_sectors; s++) {
			size += D88_SECTOR_HEADER_SIZE + track->sectors[s].sector_size;
		}
	}

	return size;
}

struct d88_track_t *get_d88_track(struct d88_disk_t *disk, int C, int H)
{
	int t;

	for (t = 0; t < D88_TRACK_COUNT; t++) {
		if (disk->tracks[t].sectors == NULL) {
			continue;
		}

		if (disk->tracks[t].sectors[0].C == C &&
			disk->tracks[t].sectors[0].H == H) {

			return &(disk->tracks[t]);
		}
	}

	return NULL;
}

struct d88_sector_t *get_d88_sector(struct d88_disk_t *disk, int C, int H, int R)
{
	int t;
	int s;

	for (t = 0; t < D88_TRACK_COUNT; t++) {
		if (disk->tracks[t].sectors == NULL) {
			continue;
		}

		if (disk->tracks[t].sectors[0].C == C &&
			disk->tracks[t].sectors[0].H == H) {

			for (s = 0; s < disk->tracks[t].num_sectors; s++) {
				if (disk->tracks[t].sectors[s].R == R) {
					return &(disk->tracks[t].sectors[s]);
				}
			}

			return NULL;
		}
	}

	return NULL;
}

/* EOF */
