#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "bitmap.h"
#include "d88image.h"


#define SHAPE_WIDTH			(16)
#define SHAPE_HEIGHT		(16)
#define NUM_SHAPES			(256)

#define SHAPE_COLS			(16)
#define SHAPE_ROWS			(16)

#define DISK_TYPE_MULTI_DISK	(-2)
#define DISK_TYPE_UNKNOWN		(-1)
#define DISK_TYPE_ULTIMA		(0)
#define DISK_TYPE_BRITANNIA		(1)
#define DISK_TYPE_TOWNE			(2)
#define DISK_TYPE_UNDERWORLD	(3)
#define DISK_TYPE_NEW_WORLD		(4)
#define NUM_DISK_TYPES			(5)


static const struct bitmap_palette_t palette[16] = {
	RGB(0x00, 0x00, 0x00),		/* black */
	RGB(0x00, 0x00, 0xFF),		/* blue */
	RGB(0xFF, 0x00, 0x00),		/* red */
	RGB(0xFF, 0x00, 0xFF),		/* magenta */
	RGB(0x00, 0xFF, 0x00),		/* green */
	RGB(0x00, 0xFF, 0xFF),		/* cyan */
	RGB(0xFF, 0xFF, 0x00),		/* yellow */
	RGB(0xFF, 0xFF, 0xFF),		/* white */
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
};

static unsigned char vram[SHAPE_WIDTH * SHAPE_COLS * SHAPE_HEIGHT * SHAPE_ROWS];

static int opt_verbose = 0;


static void usage(void);
static int read_shape_bmp(unsigned char *vram, const char *name);
static int change_shape_data(const unsigned char *vram, struct d88_disk_t *disk, int s_shp, int e_shp, int s_trk, int s_sur, int s_sec, int e_trk, int e_sur, int e_sec);
static int confirm(const char *msg);


int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 1;
	int disk_type = DISK_TYPE_UNKNOWN;
	const char *bmp_file = NULL;

	const char *image_file = NULL;
	const char *default_image_file[NUM_DISK_TYPES] = {
		"u4-ultima.d88",
		"u4-britannia.d88",
		"u4-towne.d88",
		"u4-underworld.d88",
		"u4-newworld.d88"
	};
	int image_index[NUM_DISK_TYPES] = { 0, -1, -1, -1, -1 };

	const char *new_file = NULL;
	char *tmp_name = NULL;

#if 0
	int opt_write_back = 0;
#endif
	int opt_overwrite = 0;

	struct d88_image_t image;
	struct d88_disk_t *disk = NULL;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-d") == 0 || strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("error: no image file name\n");
					return 1;
				}
				image_file = *argv;
			} else if (strcmp(*argv, "-i") == 0 || strcmp(*argv, "--index") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("error: no image index\n");
					return 1;
				}
				image_index[0] = strtoul(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0') {
					printf("error: illegal image index '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "-t") == 0 || strcmp(*argv, "--type") == 0) {
				if (*++argv == NULL) {
					printf("error: no disk type\n");
					return 1;
				}
				if (strcasecmp(*argv, "u") == 0 || strcasecmp(*argv, "ultima") == 0) {
					disk_type = DISK_TYPE_ULTIMA;
				} else if (strcasecmp(*argv, "b") == 0 || strcasecmp(*argv, "britannia") == 0) {
					disk_type = DISK_TYPE_BRITANNIA;
				} else if (strcasecmp(*argv, "t") == 0 || strcasecmp(*argv, "towne") == 0) {
					disk_type = DISK_TYPE_TOWNE;
				} else if (strcasecmp(*argv, "uw") == 0 || strcasecmp(*argv, "underworld") == 0) {
					disk_type = DISK_TYPE_UNDERWORLD;
				} else if (strcasecmp(*argv, "nw") == 0 || strcasecmp(*argv, "newworld") == 0) {
					disk_type = DISK_TYPE_NEW_WORLD;
				} else {
					printf("error: illegal disk type '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "-D") == 0 || strcmp(*argv, "--multi-disk") == 0) {
				int i;

				disk_type = DISK_TYPE_MULTI_DISK;

				if (*++argv == NULL) {
					printf("error: no disk name\n");
					return 1;
				}
				image_file = *argv;

				for (i = 0; i < NUM_DISK_TYPES; i++) {
					char *endptr;
					const char *disk_name[NUM_DISK_TYPES] = { "ultima", "britannia", "towne", "underworld", "new world" };
					if (*++argv == NULL) {
						printf("error: no image index of %s disk\n", disk_name[i]);
						return 1;
					}
					image_index[i] = strtoul(*argv, &endptr, 0);
					if (endptr == NULL || *endptr != '\0') {
						printf("error: illegal image index '%s'\n", *argv);
						return 1;
					}
				}
			} else if (strcmp(*argv, "-o") == 0 || strcmp(*argv, "--output") == 0) {
				if (*++argv == NULL) {
					printf("error: no output file name\n");
				}
				new_file = *argv;
#if 0
			} else if (strcmp(*argv, "-W") == 0 || strcmp(*argv, "--write-back") == 0) {
				opt_write_back = 1;
#endif
			} else if (strcmp(*argv, "--overwrite") == 0) {
				opt_overwrite = 1;
			} else if (strcmp(*argv, "--verbose") == 0) {
				opt_verbose = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			bmp_file = *argv;
			num_cmd_params--;
		}
	}

	if (disk_type == DISK_TYPE_UNKNOWN) {
		printf("no disk type\n");
		return 1;
	}

	if (image_file == NULL) {
		image_file = default_image_file[disk_type];
	}

	if (bmp_file == NULL) {
		printf("no shapes bitmap filename\n");
		return 1;
	}

#if 0
	if (opt_write_back) {
		new_file = image_file;
	}
#endif

	memset(vram, 0, sizeof(vram));
	memset(&image, 0, sizeof(image));

	if (read_shape_bmp(vram, bmp_file) < 0) {
		goto ERROR_EXIT;
	}

	if (read_d88_image(&image, image_file) < 0) {
		printf("error: read d88 image '%s'\n", image_file);
		goto ERROR_EXIT;
	}

	{
		int i;

		for (i = 0; i < NUM_DISK_TYPES && 0 <= image_index[i]; i++) {
			if ((disk = get_d88_disk(&image, image_index[i])) == NULL) {
				printf("error: illegal image index: '%s':%d\n", image_file, image_index[i]);
				goto ERROR_EXIT;
			}

			if (disk_type == DISK_TYPE_ULTIMA ||
				disk_type == DISK_TYPE_NEW_WORLD ||
				(disk_type == DISK_TYPE_MULTI_DISK && (i == DISK_TYPE_ULTIMA || i == DISK_TYPE_NEW_WORLD))) {
				if (change_shape_data(vram, disk,
									  0x00, 0x7F,
									  0x20, 1, 0x10,
									  0x22, 0, 0x08) < 0) {
					printf("error: change shape data\n");
					goto ERROR_EXIT;
				}
			} else {
				if (change_shape_data(vram, disk,
									  0x80, 0xFF,
									  0x1C, 1, 0x09,
									  0x1E, 1, 0x08) < 0) {
					printf("error: change shape data\n");
					goto ERROR_EXIT;
				}
			}
		}
	}

	if (new_file == NULL) {
		int len = strlen(image_file) + 5;
		char *ext;

		if ((tmp_name = (char *)malloc(len)) == NULL) {
			printf("memory exhausted!\n");
			goto ERROR_EXIT;
		}

		ext = strrchr(image_file, '.');
		if (strchr (ext, '/') != NULL) {
			ext = NULL;
		}
		len = ext - image_file;

		strncpy(tmp_name, image_file, len);
		strcpy(tmp_name + len, ".NEW");
		strcpy(tmp_name + len + 4, ext);

		new_file = tmp_name;
	}

	if (opt_overwrite == 0) {
		struct stat st;

		if (stat(new_file, &st) == 0) {
#if 0
			if (opt_write_back) {
				printf("write back to '%s'\n", new_file);
			} else {
#endif
				printf("already exists '%s'\n", new_file);
#if 0
			}
#endif

			if (confirm("overwrite ?") <= 0) {
				return -1;
			}
		}
	}

	if (write_d88_image(&image, new_file) < 0) {
		printf("error: write d88 image '%s'\n", new_file);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	free_d88_image(&image, 0);

	if (tmp_name != NULL) {
		free(tmp_name);
	}

	return result;
}

static void usage(void)
{
	printf("usage: u4pc88chg_shapes [options] SHAPES-BMP\n");
	printf("option: -d|--disk FILE    specify disk image file\n");
	printf("        -i|--index NUM    specify index of disk image\n");
	printf("        -t|--type TYPE    specify disk image type\n");
	printf("                            U |ULTIMA      ultima disk     (u4-ultima.d88)\n");
	printf("                            B |BRITANNIA   britannia disk  (u4-britannia.d88)\n");
	printf("                            T |TOWNE       towne disk      (u4-towne.d88)\n");
	printf("                            UW|UNDERWORLD  underworld disk (u4-underworld.d88)\n");
	printf("                            NW|NEWWORLD    new world disk  (u4-newworld.d88)\n");
	printf("\n");
	printf("        -D|--multi-disk FILE INDEX-U INDEX-B INDEX-T INDEX-UW INDEX-N\n");
	printf("                          modify all disk in multiple disk image\n");
	printf("\n");
	printf("        -o|--output FILE  specify modified disk image file\n");
#if 0
	printf("        -W|--write-back   write back to original file\n");
#endif
	printf("           --overwrite    overwrite output file\n");
	printf("           --verbose      verbose mode\n");
	printf("        -?|--help         print usage\n");
	printf("\n");
	printf("supported disk image format is *.D88 only\n");
	printf("write modified disk image to filename.NEW.ext if not specifyed\n");
}

static int read_shape_bmp(unsigned char *vram, const char *name)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct bitmap_palette_t *work_palette = NULL;
	unsigned char *work_data = NULL;

	if ((fp = fopen(name, "rb")) == NULL) {
		printf("error: open '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (read_bitmap(fp, &header, &work_palette, &work_data, 0) < 0) {
		printf("error: read bitmap '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (header.width != SHAPE_WIDTH * SHAPE_COLS || header.height < SHAPE_HEIGHT * SHAPE_ROWS) {
		printf("error: bmp size error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (convert_bit_count(SHAPE_WIDTH * SHAPE_COLS, SHAPE_HEIGHT * SHAPE_ROWS,
						  header.bit_count, work_palette, work_data,
						  4, palette, &vram) < 0) {
		printf("error: convert bit count '%s' %d->4\n", name, header.bit_count);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (work_palette != NULL) {
		free(work_palette);
	}
	if (work_data != NULL) {
		free(work_data);
	}

	return result;
}

static int change_shape_data(const unsigned char *vram, struct d88_disk_t *disk, int s_shp, int e_shp, int s_trk, int s_sur, int s_sec, int e_trk, int e_sur, int e_sec)
{
	int result = -1;

	int i;
	int trk;
	int sur;
	int sec;

	int plane;
	int sx;
	int sy;
	int dx;
	int dy;

	if (opt_verbose) {
		printf("%02X/%02X/%02X - %02X/%02X/%02X\n", s_trk, s_sur, s_sec, e_trk, e_sur, e_sec);
	}

	trk = s_trk;
	sur = s_sur;
	sec = s_sec;
	plane = 0;
	i = s_shp;
	sy = 0;
	sx = 0;
	dy = i / SHAPE_COLS;
	dx = i % SHAPE_COLS;

	while (i <= e_shp) {
		struct d88_sector_t *sector;
		unsigned char *des;
		int j;

		if (opt_verbose) {
			printf("[%02X/%02X/%02X]\n", trk, sur, sec);
		}

		if ((sector = get_d88_sector(disk, trk, sur, sec)) == NULL ||
			sector->N != 1) {
			printf("error: read d88 image %02X/%02X/%02X\n", trk, sur, sec);
			goto ERROR_EXIT;
		}
		des = sector->data;

		for (j = 0; j < 256 && i <= e_shp; j += (SHAPE_WIDTH / 8)) {
			const unsigned char *src = vram + (dy * SHAPE_HEIGHT + sy) * SHAPE_WIDTH * SHAPE_COLS + dx * SHAPE_WIDTH;
			int pal = (0x60 <= i && i <= 0x7E) ? 7 : (1 << plane);
			int b;
			const int mask[8] = { 0x80, 0x20, 0x08, 0x02, 0x40, 0x10, 0x04, 0x01 };

			if (opt_verbose) {
				printf("%08X :", src - vram);
			}

			for (sx = 0; sx < SHAPE_WIDTH; sx += 8) {
				int val = 0;

				for (b = 0; b < 8; b++, src++) {
					if (*src & pal) {
						val |= mask[b];
					}
				}
				*des++ = val;
				if (opt_verbose) {
					printf(" %02X", val);
				}
			}
			if (opt_verbose) {
				printf("\n");
			}

			sy++;
			if (SHAPE_HEIGHT <= sy) {
				sy = 0;
				plane++;
				if (3 <= plane || (0x60 <= i && i <= 0x7E)) {
					plane = 0;
					i++;
					dy = i / SHAPE_COLS;
					dx = i % SHAPE_COLS;
					if (0x80 <= i && (i & 3) == 0) {
						break;
					}
				}
			}
		}

		if (trk == e_trk && sur == e_sur && sec == e_sec) {
			break;
		}

		sec++;
		if (0x10 < sec) {
			sec = 0x01;
			if (sur == 0) {
				sur = 1;
			} else {
				sur = 0;
				trk++;
			}
		}
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
