#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>

#include "bitmap.h"
#include "d88image.h"


#define SHAPE_WIDTH			(16)
#define SHAPE_HEIGHT		(16)
#define NUM_SHAPES			(256)

#define SHAPE_COLS			(16)
#define SHAPE_ROWS			(16)

#define NUM_DISKS			(2)

static const struct bitmap_palette_t palette[16] = {
	RGB(0x00, 0x00, 0x00),		/* 0 black */
	RGB(0x00, 0x00, 0xFF),		/* 1 blue */
	RGB(0xFF, 0x00, 0x00),		/* 2 red */
	RGB(0xFF, 0x00, 0xFF),		/* 3 magenta */
	RGB(0x00, 0xFF, 0x00),		/* 4 green */
	RGB(0x00, 0xFF, 0xFF),		/* 5 cyan */
	RGB(0xFF, 0xFF, 0x00),		/* 6 yellow */
	RGB(0xFF, 0xFF, 0xFF),		/* 7 white */
	RGB(0x80, 0x80, 0x80),
	RGB(0x00, 0x00, 0x80),
	RGB(0x80, 0x00, 0x00),
	RGB(0x80, 0x00, 0x80),
	RGB(0x00, 0x80, 0x00),
	RGB(0x00, 0x80, 0x80),
	RGB(0x80, 0x80, 0x00),
	RGB(0xC0, 0xC0, 0xC0)
};

static unsigned char vram[SHAPE_WIDTH * SHAPE_COLS * SHAPE_HEIGHT * SHAPE_ROWS];

static int opt_verbose = 0;


static void usage(void);
static int draw_shape_data(unsigned char *vram, struct d88_disk_t *disk, int s_shp, int e_shp, int s_trk, int s_sur, int s_sec, int e_trk, int e_sur, int e_sec);
static int write_bmp_file(const char *filename, int width, int height, const unsigned char *vram, const struct bitmap_palette_t *palette, int bit_count, int opt_overwrite);
static int confirm(const char *msg);



int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 1;
	const char *bmp_file = NULL;
	const char *image_file[NUM_DISKS] = {
		"u4-ultima.d88",
		"u4-britannia.d88"
	};
	int image_index[NUM_DISKS] = { 0, 0 };

	int opt_overwrite = 0;

	struct d88_image_t image[NUM_DISKS];
	struct d88_disk_t *disk[NUM_DISKS] = { NULL, NULL };

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0 ||
				strcmp(*argv, "-?") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk1") == 0 ||
					   strcmp(*argv, "-d1") == 0) {
				if (*++argv == NULL) {
					printf("error: no disk name (ultima/new world disk)\n");
					return 1;
				}
				image_file[0] = *argv;
			} else if (strcmp(*argv, "--disk2") == 0 ||
					   strcmp(*argv, "-d2") == 0) {
				if (*++argv == NULL) {
					printf("error: no disk name (britannia/towne/underworld disk)\n");
					return 1;
				}
				image_file[1] = *argv;
			} else if (strcmp(*argv, "--index1") == 0 ||
					   strcmp(*argv, "-i1") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("error: no image index\n");
					return 1;
				}
				image_index[0] = strtoul(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0') {
					printf("error: illegal image index '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--index2") == 0 ||
					   strcmp(*argv, "-i2") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("error: no image index\n");
					return 1;
				}
				image_index[1] = strtoul(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0') {
					printf("error: illegal image index '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--multi-disk") == 0 ||
					   strcmp(*argv, "-D") == 0) {
				int i;
				if (*++argv == NULL) {
					printf("error: no disk name\n");
					return 1;
				}
				image_file[0] = image_file[1] = *argv;

				for (i = 0; i < NUM_DISKS; i++) {
					char *endptr;
					if (*++argv == NULL) {
						printf("error: no image index of %s disk\n",
							   i == 0 ? "ultima/new world" : "britannia/towne/underworld");
						return 1;
					}
					image_index[i] = strtoul(*argv, &endptr, 0);
					if (endptr == NULL || *endptr != '\0') {
						printf("error: illegal image index '%s'\n", *argv);
						return 1;
					}
				}
			} else if (strcmp(*argv, "--overwrite") == 0) {
				opt_overwrite = 1;
			} else if (strcmp(*argv, "--verbose") == 0) {
				opt_verbose = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			bmp_file = *argv;
			num_cmd_params--;
		}
	}

	if (bmp_file == NULL) {
		printf("no output bitmap filename\n");
		return 1;
	}

	memset(vram, 0, sizeof(vram));
	memset(image, 0, sizeof(image));

	/* read disk image */
	if (read_d88_image(&(image[0]), image_file[0]) < 0) {
		printf("error: read d88 image '%s'\n", image_file[0]);
		goto ERROR_EXIT;
	}

	if ((disk[0] = get_d88_disk(&(image[0]), image_index[0])) == NULL) {
		printf("error: illegal image index: '%s':%d\n", image_file[0], image_index[0]);
		goto ERROR_EXIT;
	}

	if (strcmp(image_file[0], image_file[1]) == 0) {
		if ((disk[1] = get_d88_disk(&(image[0]), image_index[1])) == NULL) {
			printf("error: illegal image index: '%s':%d\n", image_file[0], image_index[1]);
			goto ERROR_EXIT;
		}
	} else {
		if (read_d88_image(&(image[1]), image_file[1]) < 0) {
			printf("error: read d88 image '%s'\n", image_file[1]);
			goto ERROR_EXIT;
		}

		if ((disk[1] = get_d88_disk(&(image[1]), image_index[1])) == NULL) {
			printf("error: illegal image index: '%s':%d\n", image_file[1], image_index[1]);
			goto ERROR_EXIT;
		}
	}

	if (draw_shape_data(vram, disk[0],
						0x00, 0x7F,
						0x20, 1, 0x10,
						0x22, 0, 0x08) < 0 ||
		draw_shape_data(vram, disk[1],
						0x80, 0xFF,
						0x1C, 1, 0x09,
						0x1E, 1, 0x08) < 0) {
		printf("error: draw shapes\n");
		goto ERROR_EXIT;
	}

	if (write_bmp_file(bmp_file,
					   SHAPE_WIDTH * SHAPE_COLS,
					   SHAPE_HEIGHT * SHAPE_ROWS,
					   vram,
					   palette,
					   4,
					   opt_overwrite) < 0) {
		printf("error: write bitmap file '%s'\n", bmp_file);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	free_d88_image(&(image[0]), 0);
	free_d88_image(&(image[1]), 0);

	return result;
}

static void usage(void)
{
	printf("usage: u4pc88shapes [options] OUTPUT.BMP\n");
	printf("option: -d1|--disk1 FILE    specify image file of ultima/new world disk\n");
	printf("                            (default: u4-ultima.d88)\n");
	printf("        -d2|--disk2 FILE    specify image file of britannia/towne/underworld\n");
	printf("                            disk\n");
	printf("                            (default: u4-britannia.d88)\n");
	printf("        -i1|--index1 INDEX  specify index no of ultima/new world disk\n");
	printf("        -i2|--index2 INDEX  specify index no of britannia/towne/underworld disk\n");
	printf("        -D |--multi-disk FILE INDEX1 INDEX2\n");
	printf("                            same as -d1 FILE -i1 INDEX1 -d2 FILE -i2 INDEX2\n");
	printf("            --overwrite     overwrite output file\n");
	printf("            --verbose       verbose mode\n");
	printf("        -? |--help          print usage\n");
	printf("supported disk image format is *.D88 only\n");
}

static int draw_shape_data(unsigned char *vram, struct d88_disk_t *disk, int s_shp, int e_shp, int s_trk, int s_sur, int s_sec, int e_trk, int e_sur, int e_sec)
{
	int result = -1;

	int i;
	int trk;
	int sur;
	int sec;

	int plane;
	int sx;
	int sy;
	int dx;
	int dy;

	if (opt_verbose) {
		printf("%02X/%02X/%02X - %02X/%02X/%02X\n", s_trk, s_sur, s_sec, e_trk, e_sur, e_sec);
	}

	trk = s_trk;
	sur = s_sur;
	sec = s_sec;
	plane = 0;
	i = s_shp;
	sy = 0;
	sx = 0;
	dy = i / SHAPE_COLS;
	dx = i % SHAPE_COLS;

	while (i <= e_shp) {
		const struct d88_sector_t *sector;
		const unsigned char *src;
		int j;

		if (opt_verbose) {
			printf("[%02X/%02X/%02X]\n", trk, sur, sec);
		}

		if ((sector = get_d88_sector(disk, trk, sur, sec)) == NULL ||
			sector->N != 1) {
			printf("error: read d88 image %02X/%02X/%02X\n", trk, sur, sec);
			goto ERROR_EXIT;
		}
		src = sector->data;

		for (j = 0; j < 256 && i <= e_shp; j += (SHAPE_WIDTH / 8)) {
			unsigned char *des = vram + (dy * SHAPE_HEIGHT + sy) * SHAPE_WIDTH * SHAPE_COLS + dx * SHAPE_WIDTH;
			int pal = (0x60 <= i && i <= 0x7E) ? 7 : (1 << plane);
			int b;
			const int mask[8] = { 0x80, 0x20, 0x08, 0x02, 0x40, 0x10, 0x04, 0x01 };

			if (opt_verbose) {
				printf("%08X : ", des - vram);
			}

			for (sx = 0; sx < SHAPE_WIDTH; sx += 8) {
				int val = *src++;

				for (b = 0; b < 8; b++, des++) {
					if (val & mask[b]) {
						*des |= pal;

						if (opt_verbose) {
							printf("%d", pal);
						}
					} else {
						if (opt_verbose) {
							printf(" ");
						}
					}
				}
			}
			if (opt_verbose) {
				printf("\n");
			}

			sy++;
			if (SHAPE_HEIGHT <= sy) {
				sy = 0;
				plane++;
				if (3 <= plane || (0x60 <= i && i <= 0x7E)) {
					plane = 0;
					i++;
					dy = i / SHAPE_COLS;
					dx = i % SHAPE_COLS;
					if (0x80 <= i && (i & 3) == 0) {
						break;
					}
				}
			}
		}

		if (trk == e_trk && sur == e_sur && sec == e_sec) {
			break;
		}

		sec++;
		if (0x10 < sec) {
			sec = 0x01;
			if (sur == 0) {
				sur = 1;
			} else {
				sur = 0;
				trk++;
			}
		}
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static int write_bmp_file(const char *filename, int width, int height, const unsigned char *vram, const struct bitmap_palette_t *palette, int bit_count, int opt_overwrite)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;

	if (create_bitmap_header(&header, width, height, bit_count) < 0) {
		printf("error: create bitmap header\n");
		goto ERROR_EXIT;
	}

	if (opt_overwrite == 0) {
		struct stat st;

		if (stat(filename, &st) == 0) {
			printf("already exists '%s'\n", filename);
			if (confirm("overwrite ?") <= 0) {
				goto ERROR_EXIT;
			}
		}
	}

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("error: open '%s'\n", filename);
		goto ERROR_EXIT;
	}

	if (write_bitmap(fp, &header, palette, vram) < 0) {
		printf("error: write bitmap '%s'\n", filename);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
